#======
# Copyright (c) IBM Corp. 2011, 2015.
# All rights reserved. This program is made available
# under the terms of the Eclipse Public License v1.0
# which accompanies this distribution, and is available at
# http://www.eclipse.org/legal/epl-v10.html
#=======


"""Updates the IBM PowerKVM system, given an update image or repository

This program should only be used, in case your system cannot directly
access the Internet. Otherwise, you should refer to the normal yum tools

Usage: ibm-update-system [OPTIONS] --iso-path=file
       ibm-update-system [OPTIONS] --url=repository

Options:

    -h
    --help
        Print this message and exit.

    -v
    --version
        Display version information and exit.

    -y
    --yes
        Assumes yes to answers

    -s
    --iso=
        Path to iso image containing updates

    -u
    --url=
        URL (http/ftp) to iso image containing updates

    -U
    --upgrade
        Upgrade between IBM PowerKVM releases

"""

from contextlib import contextmanager
import getopt
import os
import re
import shlex
import shutil
import subprocess
import sys
from tempfile import mkdtemp
import urllib
import urllib2

import yum
from yum.constants import TS_INSTALL_STATES, TS_REMOVE_STATES
from yum import logging
from yum.rpmtrans import RPMBaseCallback

__version__ = "1.4"
__appname__ = "ibm-update-system"
SUPPORTED_UPGRADES = {"2.1.0": "2.1.1"}

#
# CONSTANTS AND DEFINITIONS
#
FOLLOW_PROGRESS = {"operation": "",
                   "percent": "",
                   "package": ""
                   }
PRODUCT_NAME = "ibm_powerkvm"
RELEASE_FILEPATH = '/etc/system-release'
REPO_VERSION_VAR_FILEPATH = '/etc/yum/vars/ibmver'
ISO_METADATA_FILENAME = '.discinfo'
ISO_ROOTFS_IMAGE_REL_FILEPATH = 'LiveOS/squashfs.img'
GPG_KEY_URL = 'file:///etc/pki/rpm-gpg/RPM-GPG-KEY-ibm_powerkvm'


def usage(code, msg=''):
    if code == 0:
        print >> sys.stderr, __doc__
    if msg:
        print >> sys.stderr, msg
    if code > 0:
        print >> sys.stderr, "Try '%s --help' for more information." % __appname__
    sys.exit(code)


def banner():
    sys.stdout.write("%s" % '='*int(len(__appname__) + len(__version__) + 1))
    sys.stdout.write("\n")
    sys.stdout.write("%s %s" % (__appname__, __version__))
    sys.stdout.write("\n")
    sys.stdout.write("%s" % '='*int(len(__appname__) + len(__version__) + 1))
    sys.stdout.write("\n")


def infoMsg(message):
    sys.stdout.write("%s\n" % message)


def runCmd(cmd, shell=False):
    """
    Run a command

    Args:
        cmd (str): command
        shell (bool): use a Linux shell

    Returns:
        str: output

    Raises:
        Exception: if return code is not 0
    """

    args = shlex.split(cmd)

    try:
        sp = subprocess.Popen(args, shell=shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    except OSError:
        raise Exception("OS error running command %s" % cmd)

    (output, err) = sp.communicate()
    rc = sp.returncode
    if rc != 0:
        raise Exception("Command %s failed, return code %s, output %s, error: %s" % (cmd, rc, output, err))

    return output


def validateInstalledOsName():
    """
    Validate installed operating system name

    Raises:
        RuntimeError: if unable to get OS name or if it is invalid
    """

    if not os.path.isfile(RELEASE_FILEPATH):
        raise RuntimeError("Unable to get installed OS name")

    with open(RELEASE_FILEPATH, 'r') as f:
        line = f.readline()
        product = line.split(' ')[0].strip().lower()
        if product.lower() != PRODUCT_NAME:
            raise RuntimeError("invalid installed OS name %s, expected was %s" % (product, PRODUCT_NAME))
        infoMsg('\tProduct name: %s' % product)

def getInstalledOsStream():
    """
    Get installed operating system stream

    Returns:
        str: OS stream

    Raises:
        RuntimeError: if unable to get OS stream
    """

    if not os.path.isfile(REPO_VERSION_VAR_FILEPATH):
        raise RuntimeError("Unable to get installed OS version")

    with open(REPO_VERSION_VAR_FILEPATH, 'r') as f:
        line = f.readline()
        version = line.rstrip('\n').strip()
        infoMsg('\tProduct base version: %s' % version)
        return version

def validateIsoName(isoPath):
    """
    Validate product name in an ISO

    Raises:
        RuntimeError: if unable to get ISO name or if it is invalid
    """

    errorMsg = "Unable to get ISO name"

    isoMetadataFilepath = os.path.join(isoPath, ISO_METADATA_FILENAME)
    if not os.path.isfile(isoMetadataFilepath):
        raise RuntimeError(errorMsg)
    with open(isoMetadataFilepath, 'r') as _file:
        # ignore first line
        _file.readline()
        line = _file.readline()
        return line.lower().strip().split(" ")[0]

    raise RuntimeError(errorMsg)

def getIsoStream(isoPath):
    """
    Get ISO stream

    Returns:
        str: ISO stream

    Raises:
        RuntimeError: if unable to get ISO stream
    """

    errorMsg = "Unable to get ISO version"

    isoMetadataFilepath = os.path.join(isoPath, ISO_METADATA_FILENAME)
    if not os.path.isfile(isoMetadataFilepath):
        raise RuntimeError(errorMsg)
    with open(isoMetadataFilepath, 'r') as _file:
        # ignore first line
        _file.readline()
        line = _file.readline()
        stream = line.lower().strip().split(" ")[1]
        if stream.startswith("e"):
            stream = stream[1:]
        return stream

    raise RuntimeError(errorMsg)

def isUpdateIso(isoPath):
    """
    Check if an ISO is an update-only ISO (only RPM packages) or if it
    a full ISO

    Args:
        isoPath (str): ISO file path

    Returns:
        bool: if ISO is an update-only ISO
    """

    isoRootFsImageFilepath = os.path.join(isoPath,
                                      ISO_ROOTFS_IMAGE_REL_FILEPATH)
    return not os.path.isfile(isoRootFsImageFilepath)

def validateIsoUpdate(osStream, isoStream, isoPath, upgrade):
    """
    Validate update/upgrade of base packages in PowerKVM ISO

    Args:
        osStream (str): installed PowerKVM stream
        isoStream (str): PowerKVM update ISO stream
        isoPath (str): PowerKVM update ISO file path
        upgrade (bool): stream update is desired

    Raises:
        RuntimeError: if validation fails
    """

    if upgrade:
        # check if ISO is full
        if isUpdateIso(isoPath):
            msg = "System can not be upgraded using an \
update ISO. Upgrade operation is only allowed when using a full ISO."
            raise RuntimeError(msg)

        # check if current version may upgrade to new version
        found = False
        for os, iso in SUPPORTED_UPGRADES.items():
            if osStream == os and isoStream == iso:
                found = True
        if not found:
            msg = "Invalid upgrade from %s to %s" % (osStream, isoStream)
            raise RuntimeError(msg)
    else:
        # check if current and new stream match
        if osStream != isoStream:
            msg = "Invalid update from %s to %s" % (osStream, isoStream)
            raise RuntimeError(msg)

def findIsoRepo(mountPoint):
    """
    Find yum repository directory in ISO

    Args:
        mountPoint (str): ISO mount point
    """

    VALID_REPO_REL_DIRS = ["repodata", "packages/repodata"]

    for repo_rel_dir in VALID_REPO_REL_DIRS:
        repo_abs_dir = os.path.join(mountPoint, repo_rel_dir)
        if os.path.exists(repo_abs_dir):
            dir_parts = repo_abs_dir.split("/")
            if dir_parts == 1:
                return mountPoint
            else:
                return os.path.join(mountPoint, "/".join(dir_parts[:-1]))

    raise RuntimeError('Could not find a valid update repository')

def parseOptions():

    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hvys:i:u:Ug',
                                   ['help', 'version', 'yes', 'iso=', 'iso-path=', 'url=',
                                    'upgrade', 'no-gpg-check'])
    except getopt.error, msg:
        usage(1, msg)

    if not opts:
        usage(1, 'No options provided')

    if args:
        usage(1, 'Invalid extra argument(s) provided')

    isoUrl = None
    assumeYes = False
    upgrade = False
    gpgCheck = True

    for opt, arg in opts:
        if opt in ('-h', '--help'):
            usage(0)
        elif opt in ('-v', '--version'):
            print >> sys.stderr, __appname__, __version__
            sys.exit(0)
        elif opt in ('-y', '--yes'):
            assumeYes = True
        elif opt in ('-s', '--iso'):
            if not arg:
                usage(1, 'Missing update ISO image path/URL')
            isoUrl = arg
        elif opt in ('-i', '--iso-path'):
            print("WARNING: -i ( --iso-path) is a deprecated option, use --iso instead")
            if not arg:
                usage(1, 'Missing update ISO image path/URL')
            if isoUrl:
                usage(1, '-s (--iso), -u (--url) and -i (--iso-path) are exclusive options and cannot be used together')
            isoUrl = arg
        elif opt in ('-u', '--url'):
            print("WARNING: -u ( --url) is a deprecated option, use --iso instead")
            if not arg:
                usage(1, 'Missing update ISO image path/URL')
            if isoUrl:
                usage(1, '-s (--iso), -u (--url) and -i (--iso-path) are exclusive options and cannot be used together')
            isoUrl = arg
        elif opt in ('-U', '--upgrade'):
            upgrade = True
        elif opt in ('-g', '--no-gpg-check'):
            gpgCheck = False

    return isoUrl, assumeYes, upgrade, gpgCheck

def main():

    isoUrl, assumeYes, mustUpgrade, gpgCheck = parseOptions()

    banner()
    if not os.getuid() == 0:
        infoMsg("Must have root privileges")
        sys.exit(1)

    tempDir = createTempDir()
    exitCode = 0

    try:

        # validate requested update
        infoMsg("Validating requested update...")
        mountPoint = mountISO(isoUrl, tempDir)
        validateInstalledOsName()
        osStream = getInstalledOsStream()
        validateIsoName(mountPoint)
        isoStream = getIsoStream(mountPoint)
        validateIsoUpdate(osStream, isoStream, mountPoint, mustUpgrade)

        # get yum repository directories
        repoPath = findIsoRepo(mountPoint)
        repoUrl = "file://" + repoPath

        # update
        infoMsg("Checking available updates...")
        with pushd(tempDir):
            if not mustUpgrade:
                update(repoUrl, tempDir, gpgCheck, assumeYes, True, False)
            else:
                upgrade(repoUrl, tempDir, gpgCheck, assumeYes,
                          osStream, isoStream)
    except Exception as e:
        msg = 'Aborting update... %s' % str(e)
        infoMsg(msg)
        exitCode = 1

    umountISO(tempDir)

    cleanup(tempDir)
    sys.exit(exitCode)


@contextmanager
def pushd(dir):
    _cwd = os.getcwd()
    os.chdir(dir)
    yield
    os.chdir(_cwd)


def downloadISO(isoUrl, targetDir):
    destFilepath = os.path.join(targetDir, isoUrl.split('/')[-1])
    infoMsg("Downloading the image: " + isoUrl)

    def dlProgress(count, blockSize, totalSize):
        percent = int(count*blockSize*100/totalSize)
        sys.stdout.write("\r" + isoUrl + "...%d%%" % percent)
        sys.stdout.flush()

    request = urllib2.Request(isoUrl)
    try:
        urllib2.urlopen(request)
    except urllib2.HTTPError:
        infoMsg('The specified server returned an error reponse.')
        raise
    except urllib2.URLError:
        infoMsg('Could not connect to the server.')
        raise

    urllib.urlretrieve(isoUrl, destFilepath, reporthook=dlProgress)
    sys.stdout.write("\n")
    sys.stdout.flush()
    return destFilepath


def mountISO(iso, targetDir):
    """
    Mount an ISO.

    Args:
        iso (str): ISO file path or URL. Downloads ISO if it is not
            available locally.
        targetDir (str): directory where ISO will be downloaded/mounted

    Raises:
        RuntimeError: if errors occur when mounting ISO
    """

    mountPoint = os.path.join(targetDir, 'iso')
    os.makedirs(mountPoint)

    URL_REGEX = "[a-z]+://.*"
    if re.match(URL_REGEX, iso):
        isoPath = downloadISO(iso, targetDir)
    else:
        isoPath = iso

    infoMsg("Mounting image...")

    if not os.path.exists(isoPath):
        raise RuntimeError("File " + isoPath + " does not exist")

    cmd = "mount -o loop %s %s" % (isoPath, mountPoint)

    try:
        runCmd(cmd)
    except:
        raise RuntimeError('Error mounting update image')

    return mountPoint

def umountISO(targetDir):
    """
    Unmount an ISO

    Args:
        targetDir (str): target directory used in mountISO()

    Raises:
        RuntimeError: if errors occur when unmounting ISO
    """

    mountPoint = os.path.join(targetDir, 'iso')
    cmd = "umount %s" % mountPoint

    try:
        runCmd(cmd)
    except:
        raise RuntimeError('Error umounting update image')


def createTempDir():
    tempDir = mkdtemp('', 'ibmupd-', '/tmp')
    return tempDir


def cleanup(tempDir):

    shutil.rmtree(tempDir)


def confirmInstallation():
    """
    Confirm if user wants to proceed with installation
    """

    answer = None

    while answer is None:
        answer = raw_input("\nDo you want to proceed? (y/N) ")

        if answer.lower() == "y" or answer.lower() == "yes":
            answer = True
        elif answer.lower() == "n" or answer.lower() == "no":
            answer = False
        else:
            answer = None

    return answer


def createYumHandler():
    """
    Create a yum handler

    Returns:
        YumBase: yum handler
    """

    # instantiate yum main object
    yumHandler = yum.YumBase()

    # disable all repositories
    yumHandler.repos.disableRepo('*')

    return yumHandler


def closeYumHandler(yh):
    """
    Close a yum handler

    Args:
        yh (YumBase): yum handler
    """

    # close yum instance
    yh.closeRpmDB()
    yh.close()


def createYumLogger():

    # create yum logger
    yumLogger = yum.logging.getLogger("yum.verbose.YumBase")
    yumLogger.setLevel(logging.CRITICAL)
    return yumLogger

def diffLists(*lists):

    i = 0
    result = set(lists[0])
    for i in range(1, len(lists)):
        result -= set(lists[i])

    return list(result)

def fromYumPackagesObjsListToDict(packagesObjs):

    packagesDicts = []
    for pkg in packagesObjs:
        pkg_nevra =  {
          "name": pkg.name,
          "epoch": pkg.epoch,
          "version": pkg.version,
          "release": pkg.release,
          "arch": pkg.arch }
        packagesDicts.append(pkg_nevra)

    return packagesDicts

def fromRpmPackagesTuplesListToDict(packagesTuples):

    packagesDicts = []
    for pkg in packagesTuples:
        pkg_nevra =  {
          "name": pkg[0],
          "epoch": pkg[1],
          "version": pkg[2],
          "release": pkg[3],
          "arch": pkg[4] }
        packagesDicts.append(pkg_nevra)

    return packagesDicts

def printPackages(packages):

    packagesStrings = []
    for pkg in packages:
        packagesStrings.append(yum.YumBase._ui_nevra_dict(pkg))
    return "\n".join(packagesStrings)


def addRepo(yh, repoName, repoUrl, tempDir, gpgCheck, markAll=True):
    """
    Add a yum repository to a yum handler

    Args:
        yh (YumBase): yum handler
        repoName (str): yum repository name
        repoUrl (str): yum repository URL
        tempDir (str): temporary directory used to cache rpm packages
        gpgCheck (bool): check GPG signature in rpm packages
        markAll (bool):
    """

    if yh is None:
        raise RuntimeError('Invalid YumBase object.')

    basecache = os.path.join(tempDir, 'yum')
    if not os.path.isdir(basecache):
        os.makedirs(basecache)

    # configure yum parameters
    yh.conf.gpgcheck = gpgCheck
    yh.conf.assumeyes = True
    yh.conf.cache = 0
    yh.conf.basecachedir = basecache

    # create a repo
    newrepo = yum.yumRepo.YumRepository(repoName)
    newrepo.name = repoName
    newrepo.baseurl = [repoUrl]
    newrepo.basecachedir = yh.conf.basecachedir
    newrepo.enablegroups = True
    newrepo.gpgcheck = gpgCheck
    if gpgCheck:
        newrepo.gpgkey = GPG_KEY_URL
    yh.repos.add(newrepo)
    yh.repos.enableRepo(newrepo.id)
    yh.doRepoSetup(thisrepo=newrepo.id)

    try:
        yh.pkgSack
    except RuntimeError as e:
        raise RuntimeError('Error creating yum instance: %s' % e)

    if markAll:
        # mark all packages to install
        yh.repos.populateSack()
        yh.pkgSack.buildIndexes()

    return yh

def getUpdatePackages(yh):
    """
    Get list of names of update packages in yum repositories.
    In this context, "update" is a package which already has some
        version installed in the system

    Args:
        yh (YumBase): yum handler

    Returns:
        [dict]: list of updates packages dicts. Dict keys: name,
            epoch, version, release, arch
    """

    if yh is None:
        raise RuntimeError('Invalid YumBase object.')

    updatesObjs = yh.doPackageLists('updates').updates
    updatesObjs = sorted(updatesObjs)
    updatesDicts = fromYumPackagesObjsListToDict(updatesObjs)
    return updatesDicts

def getNewPackages(yh):
    """
    Get list of names of new packages in yum repositories.
    In this context, "new" is a package which has no version installed
        in the system

    Args:
        yh (YumBase): yum handler

    Returns:
        [dict]: list of new packages dicts. Dict keys: name, epoch,
            version, release, arch
    """

    if yh is None:
        raise RuntimeError('Invalid YumBase object.')

    # get lists of installed, available (in repositories) and updates packages
    # tricky: doPackageLists("all") returns only installed and available packages, so
    #         running it once to get updates, installed and available will not work
    updatePackages = yh.doPackageLists("updates").updates
    installedPackages = yh.doPackageLists("installed").installed
    availablePackages = yh.doPackageLists("available").available

    # generate lists of new packages
    newPackages = diffLists(availablePackages, installedPackages)
    if updatePackages:
        newPackages = diffLists(newPackages, updatePackages)
    newPackages = sorted(newPackages)
    newPackagesDicts = fromYumPackagesObjsListToDict(newPackages)

    return newPackagesDicts


def getObsoletePackages(yh):
    """
    Get list of names of installed obsolete packages
    In rpm terminology, an obsolete package is an installed package
        which was renamed in the yum repository.

    Args:
        yh (YumBase): yum handler

    Returns:
        [dict]: list of obsoletes packages dicts. Dict keys: name,
            epoch, version, release, arch
    """

    if yh is None:
        raise RuntimeError('Invalid YumBase object.')

    updateObj = yh.up
    obsoletesTuples = updateObj.getObsoletersTuples()
    obsoletesTuples2 = []
    for _tuple in obsoletesTuples:
        obsoletesTuples2.append(obsoletesTuples[1])
    obsoletePackagesDicts = fromRpmPackagesTuplesListToDict(obsoletesTuples2)
    return obsoletePackagesDicts

def executeYumTransaction(yh, testMode):
    """
    Execute a yum transaction of installing rpm packages

    Args:
        yh (YumBase): yum handler
        testMode (bool): if package installation must be executed
    """

    if yh is None:
        raise RuntimeError('Invalid YumBase object.')

    # Resolve package dependencies
    (rescode, restring) = yh.resolveDeps()
    if rescode == 1:
        errorMsg = "Error while resolving dependency."
        if len(restring) >= 1:
            errorMsg = str(restring)
        raise Exception(errorMsg)

    # Build yum transaction
    rescode, restring = yh.buildTransaction()
    if rescode == 1:
        errorMsg = "Error while building transaction."
        if len(restring) >= 1:
            errorMsg = str(restring)
        raise Exception(errorMsg)

    # Process yum transaction
    if not testMode:
        yh.processTransaction(
            None,
            None,
            UpdateCallback())
        print "\n"

def update(repoUrl, tempDir, gpgCheck, assumeYes=True, verbose=True,
           testMode=False):
    """
    Update packages in a system

    Args:
        repoUrl (str): yum repository URL
        tempDir (str): temporary directory used to cache rpm packages
        assumeYes (bool): if package update confirmation may be assumed
        verbose (bool): verbose output
        testMode (bool): if package installation must be executed
    """

    proceedInstallation = False

    # create yum handler
    yh = createYumHandler()
    ylg = createYumLogger()

    # add repositories
    yh = addRepo(yh, "base", repoUrl, tempDir, gpgCheck)

    # Get list of new/update/obsoletes packages
    newPackages = getNewPackages(yh)
    updatePackages = getUpdatePackages(yh)
    obsoletePackages = getObsoletePackages(yh)

    def packagesInGroup(pkgsDicts, group):
        """
        Returns a list of packages that are members of the current group.

        Args:
            pkgsDicts ([dict]): list of packages dictionaries
            group (dict): group dictionary
        Returns:
            [dict]: list of selected packages dictionaries
        """
        return [pkg for pkg in pkgsDicts if pkg['name'] in group.packages]

    if verbose:
        infoMsg('================================================================================')
        infoMsg('Packages groups')
        infoMsg('================================================================================')


    installedGroups, _ = yh.doGroupLists()
    packagesChanges = False

    for group in installedGroups:
        groupNewPkgs = packagesInGroup(newPackages, group)
        groupUpdatePkgs = packagesInGroup(updatePackages, group)
        groupObsoletePkgs= packagesInGroup(obsoletePackages, group)

        # Check if there are changes in the packages of this group
        if groupNewPkgs or groupUpdatePkgs or groupObsoletePkgs:
            packagesChanges = True
            yh.selectGroup(group.groupid)

            if verbose:
                print("\nGroup '%s'" % group.name)

                if groupUpdatePkgs:
                    print("Updating %d packages:\n%s\n" % (len(groupUpdatePkgs),
                                                           printPackages(groupUpdatePkgs)))
                if groupObsoletePkgs:
                    print("Obsoleting %d packages:\n%s\n" % (len(groupObsoletePkgs),
                                                             printPackages(groupObsoletePkgs)))
                if groupNewPkgs:
                    print("Installing %d new packages:\n%s\n" % (len(groupNewPkgs),
                                                                 printPackages(groupNewPkgs)))
    if packagesChanges:
        if not assumeYes:
            proceedInstallation = confirmInstallation()
        else:
            proceedInstallation = True
        if proceedInstallation:
            executeYumTransaction(yh, testMode)
    else:
        if verbose:
            print 'No package updates found.'

    closeYumHandler(yh)

    return proceedInstallation

def upgrade(repoUrl, tempDir, gpgCheck, assumeYes, currentStream, newStream):
    # test upgrade befora proceed
    userInput = update(repoUrl, tempDir, assumeYes, True, True)

    if userInput:
        yh = createYumHandler()
        ylg = createYumLogger()
        yh = addRepo(yh, "base", repoUrl, tempDir, gpgCheck, False)

        # mark to pkvm2_1_1-release to update
        searchlist = ['name']
        args = ["pkvm2_1_1-release"]
        matching = yh.searchGenerator(searchlist, args)
        for (po, matched_value) in matching:
            if po.name == "pkvm2_1_1-release":
                try:
                    yh.update(po)
                except:
                    pass

        # build and process the batch transaction
        executeYumTransaction(yh, False)

        closeYumHandler(yh)

        # execute update
        update(repoUrl, tempDir, True, False, False)

class UpdateCallback(RPMBaseCallback):
    def __init__(self):
        """
        Constructor
        """
        # call parent constructor
        RPMBaseCallback.__init__(self)

        # create start lock
        self.__startLock = {}
        self.__lastpackage = ""

    def errorlog(self, msg):
        """
        Dummy method. Just ignores any error messages sent from the RPM
            installation process.

        Args:
            msg (basestring): RPM installation error messages
        """
        pass

    def event(self, package, action, bytesProcessed, bytesTotal, currentProcess, totalProcess):
        """
        Handles installation progress

        Args:
            package (basestring): A yum package object or simple string
                of a package name
            action (int): A yum.constant transaction set state or in
               the obscure rpm repackage case it could be the string
               'repackaging'
            bytesProcessed (int): Current number of bytes processed in
                the transaction element being processed
            bytesTotal (int): Total number of bytes in the transaction
                element being processed
            currentProcess (int): number of processes completed in
                whole transaction
            totalProcess (int): total number of processes in the
                transaction.
        """
        # action is install: get package name and save state
        if action in TS_INSTALL_STATES:
            packageSplited = package._remote_url().split('/')
            packageName = packageSplited[len(packageSplited) - 1]
            state = 'Processing'

        # state is uninstall: get package name and save state
        elif action in TS_REMOVE_STATES:
            packageName = package
            state = 'Cleanup'

        # other action: ignore
        else:
            return

        # calculate percentage
        percentage = (float(bytesProcessed) / float(bytesTotal)) * 100

        # create message
        message = FOLLOW_PROGRESS

        # lock not exists: operation is starting
        if self.__startLock.get(packageName) is None:
            message['operation'] = '%s-start' % state
            self.__startLock[packageName] = True

        # percentage is 100: operation finish
        elif percentage == 100:
            message['operation'] = '%s-finish' % state

        # operation is running
        else:
            message['operation'] = '%s' % state

        # fill message dict
        message['percent'] = percentage
        message['package'] = packageName

        try:
            # break line when package change
            if self.__lastpackage != packageName:
                self.__lastpackage = packageName
                sys.stdout.write('\n')

            sys.stdout.write("\r%s (%s) %d%% [%d/%d]" % (state, packageName, percentage, currentProcess, totalProcess))
            sys.stdout.flush()

        except:
            print percentage
            pass


    def filelog(self, package, action):
        """
        Dummy method. Just ignores any log messages sent from the RPM
        installation process.

        Args:
            package (basestring): package name
            action (basestring): RPM installation log message
        """
        pass

    def scriptout(self, package, msgs):
        """
        Dummy method. Just ignores any scriptlet output messages sent
            from the RPM installation process.

        Args:
            package (basestring): package name
            msgs (basestring): RPM scriptlet output messages
        """
        pass


if __name__ == '__main__':
    main()
